#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
from os_vm_expire.common import utils


def convert_resource_id_to_href(resource_slug, resource_id):
    """Convert the resource ID to a HATEOAS-style href with resource slug."""
    if resource_id:
        resource = '{slug}/{id}'.format(slug=resource_slug, id=resource_id)
    else:
        resource = '{slug}/????'.format(slug=resource_slug)
    return utils.hostname_for_refs(resource=resource)


def convert_vmexpire_to_href(vmexpire_id):
    """Convert the secret-store ID to a HATEOAS-style href."""
    return convert_resource_id_to_href('vmexpire', vmexpire_id)


def convert_to_hrefs(fields):
    """Convert id's within a fields dict to HATEOAS-style hrefs."""
    url = convert_vmexpire_to_href(fields['id'])
    fields['links'] = {'self': url}
    return fields


def convert_list_to_href(resources_name, offset, limit):
    """Supports pretty output of paged-list hrefs.

    Convert the offset/limit info to a HATEOAS-style href
    suitable for use in a list navigation paging interface.
    """
    resource = '{0}?limit={1}&offset={2}'.format(resources_name, limit,
                                                 offset)
    return utils.hostname_for_refs(resource=resource)


def previous_href(resources_name, offset, limit):
    """Supports pretty output of previous-page hrefs.

    Create a HATEOAS-style 'previous' href suitable for use in a list
    navigation paging interface, assuming the provided values are the
    currently viewed page.
    """
    offset = max(0, offset - limit)
    return convert_list_to_href(resources_name, offset, limit)


def next_href(resources_name, offset, limit):
    """Supports pretty output of next-page hrefs.

    Create a HATEOAS-style 'next' href suitable for use in a list
    navigation paging interface, assuming the provided values are the
    currently viewed page.
    """
    offset = offset + limit
    return convert_list_to_href(resources_name, offset, limit)


def add_nav_hrefs(resources_name, offset, limit,
                  total_elements, data):
    """Adds next and/or previous hrefs to paged list responses.

    :param resources_name: Name of api resource
    :param offset: Element number (ie. index) where current page starts
    :param limit: Max amount of elements listed on current page
    :param total_elements: Total number of elements
    :returns: augmented dictionary with next and/or previous hrefs
    """
    if offset > 0:
        data.update({'previous': previous_href(resources_name,
                                               offset,
                                               limit)})
    if total_elements > (offset + limit):
        data.update({'next': next_href(resources_name,
                                       offset,
                                       limit)})
    return data


def add_self_href(resource_name, data):
    """Add self href to response

    :param resources_name: Name of api resource
    :returns: augmented dictionary with next and/or previous hrefs
    """
    if 'links' not in data:
        data['links'] = {}
    data['links'].update({'self': utils.hostname_for_refs(resource=resource_name)})
    return data


def get_vmexpire_id_from_ref(vmexpire_ref):
    """Parse a container reference and return the container ID

    The container ID is the right-most element of the URL
    :param container_ref: HTTP reference of container
    :return: a string containing the ID of the container
    """
    vmexpire_id = vmexpire_ref.rsplit('/', 1)[1]
    return vmexpire_id



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_policy import policy


rules = [
    policy.RuleDefault('context_is_admin',
                       'role:admin'),
    policy.RuleDefault('admin_or_owner',
                       'rule:context_is_admin or is_admin:True or project_id:%(project_id)s'),
    policy.RuleDefault('default',
                       'rule:admin_or_owner'),
]


{
    "context_is_admin": "role:admin",
    "admin_or_owner": "rule:context_is_admin or is_admin:True or project_id:%(project_id)s",
    "default": "rule:admin_or_owner",
    "vmexpire:get": "rule:admin_or_owner",
    "vmexpire:extend": "rule:admin_or_owner",
    "vmexpire:delete": "rule:context_is_admin",
    "vmexclude:get": "rule:admin",
    "vmexclude:create": "rule:admin",
    "vmexclude:delete": "rule:admin"
}


def list_rules():
    return rules



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import mock
import time

from os_vm_expire.cmd.cleaner import check as cleaner_check
from os_vm_expire.model import models
from os_vm_expire.model import repositories
# from os_vm_expire.cmd.cleaner import send_email as cleaner_send_email
from os_vm_expire.tests import utils


class MockResponse(object):
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code
        self.headers = {
            'X-Subject-Token': 'XXX'
        }

    def json(self):
        return self.json_data


def mocked_requests_get(*args, **kwargs):
    return MockResponse(None, 404)


def mocked_requests_post(*args, **kwargs):
    return MockResponse(None, 404)


def mocked_email(*args, **kwargs):
    return True


def mocked_delete_vm(instance_id, project_id, token):
    return True


class WhenTestingVmExpiresResource(utils.OsVMExpireAPIBaseTestCase):

    def setUp(self):
        super(WhenTestingVmExpiresResource, self).setUp()

    def tearDown(self):
        super(WhenTestingVmExpiresResource, self).tearDown()
        repo = repositories.get_vmexpire_repository()
        repo.delete_all_entities()
        repositories.commit()

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    @mock.patch('os_vm_expire.cmd.cleaner.send_email', side_effect=mocked_email)
    def test_vm_expire_not_cleaned(self, mock_get, mock_post, mock_email):
        entity = create_vmexpire_model('12345')
        create_vmexpire(entity)
        cleaner_check(None)
        db_entity = get_vmexpire(entity.id)
        self.assertTrue(db_entity.id == entity.id)

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    @mock.patch('os_vm_expire.cmd.cleaner.delete_vm', side_effect=mocked_delete_vm)
    @mock.patch('os_vm_expire.cmd.cleaner.send_email', side_effect=mocked_email)
    def test_vm_expire_cleaned(self, mock_get, mock_post, mock_delete, mock_email):
        entity = create_vmexpire_model('12345')
        instance = create_vmexpire(entity)
        instance.expire = 1
        instance.save()
        repositories.commit()
        cleaner_check(None)
        db_entity = get_vmexpire(entity.id)
        self.assertTrue(db_entity.notified)
        self.assertFalse(db_entity.notified_last)
        cleaner_check(None)
        self.assertTrue(db_entity.notified)
        self.assertTrue(db_entity.notified_last)
        cleaner_check(None)
        found = False
        try:
            db_entity = get_vmexpire(entity.id)
            found = True
        except Exception:
            found = False
        self.assertFalse(found)


def create_vmexpire_model(prefix=None):
    if not prefix:
        prefix = '12345'
    entity = models.VmExpire()
    entity.user_id = prefix + 'user'
    entity.project_id = prefix + 'project'
    entity.instance_id = prefix + 'instance'
    entity.instance_name = prefix
    entity.notified = False
    entity.notified_last = False
    entity.expire = time.mktime(datetime.datetime.now().timetuple())
    return entity


def create_vmexpire(entity):
    repo = repositories.get_vmexpire_repository()
    instance = repo.create_from(entity)
    repositories.commit()
    return instance


def get_vmexpire(entity_id):
    repo = repositories.get_vmexpire_repository()
    instance = repo.get(entity_id=entity_id)
    return instance



# Copyright (c) 2013-2014 Rackspace, Inc.
#               2017 IRISA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Server-side (i.e. worker side) classes and logic.
"""
import datetime
import functools
import json
import requests
import time

import oslo_messaging

from oslo_service import service

from os_vm_expire.common import config
from os_vm_expire.common import utils
from os_vm_expire.model import models
from os_vm_expire.model import repositories

CONF = config.CONF

LOG = utils.getLogger(__name__)


def get_identity_token():
    conf_worker = config.CONF.worker
    ks_uri = conf_worker.auth_uri

    auth = {
        'auth': {
            'scope':
                {'project': {
                    'name': conf_worker.admin_service,
                    'domain':
                        {
                            'name': conf_worker.admin_project_domain_name
                        }
                    }
                 },
            'identity': {
                    'password': {
                        'user': {
                            'domain': {
                                'name': conf_worker.admin_user_domain_name
                            },
                            'password': conf_worker.admin_password,
                            'name': conf_worker.admin_user
                        }
                    },
                    'methods': ['password']
                }
        }
    }
    r = requests.post(ks_uri + '/auth/tokens', json=auth)
    if 'X-Subject-Token' not in r.headers:
        LOG.error('Could not get authorization')
        return None
    token = r.headers['X-Subject-Token']
    return token


def get_project_domain(project_id):
    token = get_identity_token()
    if not token:
        return None
    conf_worker = config.CONF.worker
    ks_uri = conf_worker.auth_uri
    headers = {
        'X-Auth-Token': token,
        'Content-Type': 'application/json'
    }
    r = requests.get(ks_uri + '/projects/' + str(project_id), headers=headers)
    if not r.status_code == 200:
        LOG.error('Failed to get domain_id for project ' + str(project_id))
        return None
    project = r.json()
    domain_id = project['project']['domain_id']
    return domain_id


def find_function_name(func, if_no_name=None):
    """Returns pretty-formatted function name."""
    return getattr(func, '__name__', if_no_name)


def transactional(fn):
    """Provides request-scoped database transaction support to tasks."""

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        fn_name = find_function_name(fn, if_no_name='???')

        # Manage session/transaction.
        try:
            fn(*args, **kwargs)
            repositories.commit()
            LOG.debug("Completed worker task (post-commit): '%s'", fn_name)
        except Exception:
            """NOTE: Wrapped functions must process with care!
            Exceptions that reach here will revert the entire transaction,
            including any updates made to entities such as setting error
            codes and error messages.
            """
            LOG.exception("Problem seen processing worker task: '%s'",
                          fn_name
                          )
            repositories.rollback()
        finally:
            repositories.clear()

    return wrapper


def monitored(fn):  # pragma: no cover
    """Provides monitoring capabilities for task methods."""

    return fn


class Tasks(object):
    """Tasks that can be invoked asynchronously.

    Only place task methods and implementations on this class, as they can be
    called directly from the client side for non-asynchronous standalone
    single-node operation.
    If a new method is added that can be retried, please also add its method
    name to MAP_RETRY_TASKS above.
    The TaskServer class below extends this class to implement a worker-side
    server utilizing Oslo messaging's RPC server. This RPC server can invoke
    methods on itself, which include the methods in this class.
    """

    @monitored
    @transactional
    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        if event_type == 'instance.create.end':
            LOG.debug(event_type + ':' + payload['nova_object.data']['uuid'])
            repo = repositories.get_vmexpire_repository()
            instance = None
            instance_uuid = str(payload['nova_object.data']['uuid'])
            try:
                instance = repo.get_by_instance(instance_uuid)
            except Exception:
                LOG.debug("Fine, instance does not already exists")
            if instance:
                LOG.warn("InstanceAlreadyExists:" +
                         instance_uuid +
                         ", deleting first"
                         )
                repo.delete_entity_by_id(entity_id=instance.id)
            entity = models.VmExpire()
            entity.instance_id = instance_uuid
            if payload['nova_object.data']['display_name']:
                display_name = payload['nova_object.data']['display_name']
                entity.instance_name = display_name
            else:
                entity.instance_name = instance_uuid
            entity.project_id = payload['nova_object.data']['tenant_id']
            entity.user_id = payload['nova_object.data']['user_id']
            entity.expire = int(
                time.mktime(datetime.datetime.now().timetuple()) +
                (CONF.max_vm_duration * 3600 * 24)
                )
            entity.notified = False
            entity.notified_last = False

            project_domain = None
            try:
                project_domain = get_project_domain(entity.project_id)
            except Exception:
                LOG.exception('Failed to get domain for project')

            exclude_repo = repositories.get_vmexclude_repository()
            if project_domain:
                exclude_id = exclude_repo.get_exclude_by_id(project_domain)
                if exclude_id:
                    LOG.debug('domain %s is excluded, skipping' % (project_domain))
                    return
            exclude_id = exclude_repo.get_exclude_by_id(entity.project_id)
            if exclude_id:
                LOG.debug('project %s is excluded, skipping' % (entity.project_id))
                return
            exclude_id = exclude_repo.get_exclude_by_id(entity.user_id)
            if exclude_id:
                LOG.debug('user %s is excluded, skipping' % (entity.user_id))
                return

            instance = repo.create_from(entity)
            LOG.debug("NewInstanceExpiration:" + instance_uuid)
        elif event_type == 'instance.delete.end':
            instance_uuid = str(payload['nova_object.data']['uuid'])
            LOG.debug(event_type + ':' + instance_uuid)
            repo = repositories.get_vmexpire_repository()
            try:
                instance = repo.get_by_instance(instance_uuid)
                repo.delete_entity_by_id(entity_id=instance.id)
                LOG.debug("Delete id:" + instance.id)
            except Exception:
                LOG.warn('Failed to delete: ' + instance_uuid)

        LOG.debug(publisher_id)
        LOG.debug(event_type)
        LOG.debug(str(metadata))
        LOG.debug(json.dumps(payload, indent=4))

    def warn(self, ctxt, publisher_id, event_type, payload, metadata):
        LOG.debug(json.dumps(payload, indent=4))

    def error(self, ctxt, publisher_id, event_type, payload, metadata):
        LOG.debug(json.dumps(payload, indent=4))


class TaskServer(Tasks, service.Service):
    """Server to process asynchronous tasking from API nodes.

    This server is an Oslo service that exposes task methods that can
    be invoked from the Barbican API nodes. It delegates to an Oslo
    RPC messaging server to invoke methods asynchronously on this class.
    Since this class also extends the Tasks class above, its task-based
    methods are hence available to the RPC messaging server.
    """
    def __init__(self):
        super(TaskServer, self).__init__()

        # Setting up db engine to avoid lazy initialization
        repositories.setup_database_engine_and_factory()

        transport = oslo_messaging.get_transport(CONF)

        conf_opts = getattr(CONF, config.KS_NOTIFICATIONS_GRP_NAME)
        targets = [
            oslo_messaging.Target(
                topic=conf_opts.topic,
                exchange=conf_opts.control_exchange
                )
        ]
        endpoints = [self]
        self._server = oslo_messaging.get_notification_listener(
            transport,
            targets,
            endpoints,
            pool=conf_opts.pool_name
            )

    def start(self):
        LOG.info("Starting the TaskServer")
        self._server.start()
        super(TaskServer, self).start()

    def stop(self):
        LOG.info("Halting the TaskServer")
        super(TaskServer, self).stop()
        self._server.stop()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Supports database/repositories oriented unit testing.
Warning: Do not merge this content with the utils.py module, as doing so will
break the DevStack functional test discovery process.
"""
import datetime
import time

# from sqlalchemy.engine import Engine
# from sqlalchemy import event

from os_vm_expire.model.migration import commands
from os_vm_expire.model import models
from os_vm_expire.model import repositories
from oslo_db import options
import oslotest.base as oslotest


def create_vmexpire(instance=None, session=None):
    expire = models.VmExpire()
    expire.project_id = instance.project_id
    expire.instance_id = instance.instance_id
    expire.instance_name = instance.instance_name
    expire.user_id = instance.user_id
    expire.expire = time.mktime(datetime.datetime.now().timetuple()) + 2 * 3600 * 24
    expire.notified = False
    expire.notified_last = False
    container_repo = repositories.get_vmexpire_repository()
    container_repo.create_from(expire, session=session)
    return expire


def create_vmexclude(exclude_id=None, exclude_type=0, session=None):
    exclude = models.VmExclude()
    exclude.exclude_id = exclude_id
    exclude.exclude_type = exclude_type
    container_repo = repositories.get_vmexclude_repository()
    container_repo.create_exclude(exclude, session=session)
    return exclude


def setup_in_memory_db():
    # Ensure we are using in-memory SQLite database, and creating tables.
    options.set_defaults(repositories.CONF, connection='sqlite:///test.db')

    repositories.CONF.set_override("debug", True)
    config = commands.init_config()
    config.osvmexpire = repositories.CONF

    # Ensure the connection is completely closed, so any previous in-memory
    # database can be removed prior to starting the next test run.
    repositories.hard_reset()
    commands.upgrade(config=config, to_version='head')
    # Start the in-memory database, creating required tables.
    repositories.start()


def in_memory_cleanup():
    repositories.clear()


def get_session():
    return repositories.get_session()


class RepositoryTestCase(oslotest.BaseTestCase):
    """Base test case class for in-memory database unit tests.

    Database/Repository oriented unit tests should *not* modify the global
    state in the os_vm_expire/model/repositories.py module, as this can lead to
    hard to debug errors. Instead only utilize methods in this fixture.
    Also, database-oriented unit tests extending this class MUST NO INVOKE
    the repositories.start()/clear()/hard_reset() methods!*, otherwise *VERY*
    hard to debug 'Broken Pipe' errors could result!
    """
    def setUp(self):
        super(RepositoryTestCase, self).setUp()
        setup_in_memory_db()

        # Clean up once tests are completed.
        self.addCleanup(in_memory_cleanup)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import mock
# import sqlalchemy.orm as sa_orm
import time

from os_vm_expire.model import models
from os_vm_expire.model import repositories
from os_vm_expire.queue.server import Tasks
from os_vm_expire.tests import utils


class MockResponse(object):
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code
        self.headers = {
            'X-Subject-Token': 'XXX'
        }

    def json(self):
        return self.json_data


def mocked_requests_get(*args, **kwargs):
    return MockResponse(None, 404)


def mocked_requests_post(*args, **kwargs):
    return MockResponse(None, 404)


def mocked_get_project_domain(project_id):
    return '12345domain'


class WhenTestingVmExpiresResource(utils.OsVMExpireAPIBaseTestCase):

    def setUp(self):
        super(WhenTestingVmExpiresResource, self).setUp()
        self.task = Tasks()

    def tearDown(self):
        super(WhenTestingVmExpiresResource, self).tearDown()
        repo = repositories.get_vmexpire_repository()
        repo.delete_all_entities()
        repositories.commit()
        exclude_repo = repositories.get_vmexclude_repository()
        exclude_repo.delete_all_entities()
        repositories.commit()

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    @mock.patch('os_vm_expire.queue.server.get_project_domain', side_effect=mocked_get_project_domain)
    def test_vm_create(self, mock_get, mock_post, mock_get_project_domain):
        create_msg = {
            'nova_object.data': {
                'uuid': '1-2-3-4-5',
                'display_name': '12345',
                'tenant_id': '12345project',
                'user_id': '12345user'
            }
        }
        self.task.info(None, 'mock', 'instance.create.end', create_msg, None)
        repo = repositories.get_vmexpire_repository()
        expire = repo.get_by_instance(create_msg['nova_object.data']['uuid'])
        self.assertEqual(expire.instance_id, create_msg['nova_object.data']['uuid'])
        self.assertTrue(expire.expire > 0)

    def test_vm_delete(self):
        self.test_vm_create()
        delete_msg = {
            'nova_object.data': {
                'uuid': '1-2-3-4-5',
                'display_name': '12345',
                'tenant_id': '12345project',
                'user_id': '12345user'
            }
        }
        self.task.info(None, 'mock', 'instance.delete.end', delete_msg, None)
        repo = repositories.get_vmexpire_repository()
        found = False
        try:
            expire = repo.get_by_instance(delete_msg['nova_object.data']['uuid'])
            if expire:
                found = True
        except Exception as e:
            logging.exception(e)
            found = False
        self.assertFalse(found)


class WhenTestingVmExcludesResource(utils.OsVMExpireAPIBaseTestCase):

    def setUp(self):
        super(WhenTestingVmExcludesResource, self).setUp()
        self.task = Tasks()

    def tearDown(self):
        super(WhenTestingVmExcludesResource, self).tearDown()
        repo = repositories.get_vmexpire_repository()
        repo.delete_all_entities()
        repositories.commit()
        exclude_repo = repositories.get_vmexclude_repository()
        exclude_repo.delete_all_entities()
        repositories.commit()

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    @mock.patch('os_vm_expire.queue.server.get_project_domain', side_effect=mocked_get_project_domain)
    def test_vm_exclude_domain(self, mock_get, mock_post, mock_get_project_domain):
        create_msg = {
            'nova_object.data': {
                'uuid': '1-2-3-4-5',
                'display_name': '12345',
                'tenant_id': '12345project',
                'user_id': '12345user'
            }
        }
        entity = create_vmexclude_model('12345domain', 0)
        create_vmexclude(entity)
        self.task.info(None, 'mock', 'instance.create.end', create_msg, None)
        repo = repositories.get_vmexpire_repository()
        try:
            repo.get_by_instance(create_msg['nova_object.data']['uuid'])
        except Exception as e:
            logging.info('instance not found: ' + str(e))
        else:
            self.self.fail('domain is excluded, should not have been created')

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    @mock.patch('os_vm_expire.queue.server.get_project_domain', side_effect=mocked_get_project_domain)
    def test_vm_exclude_project(self, mock_get, mock_post, mock_get_project_domain):
        create_msg = {
            'nova_object.data': {
                'uuid': '1-2-3-4-5',
                'display_name': '12345',
                'tenant_id': '12345project',
                'user_id': '12345user'
            }
        }
        entity = create_vmexclude_model('12345project', 1)
        create_vmexclude(entity)
        self.task.info(None, 'mock', 'instance.create.end', create_msg, None)
        repo = repositories.get_vmexpire_repository()
        try:
            repo.get_by_instance(create_msg['nova_object.data']['uuid'])
        except Exception as e:
            logging.info('instance not found: ' + str(e))
        else:
            self.self.fail('domain is excluded, should not have been created')

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    @mock.patch('os_vm_expire.queue.server.get_project_domain', side_effect=mocked_get_project_domain)
    def test_vm_exclude_user(self, mock_get, mock_post, mock_get_project_domain):
        create_msg = {
            'nova_object.data': {
                'uuid': '1-2-3-4-5',
                'display_name': '12345',
                'tenant_id': '12345project',
                'user_id': '12345user'
            }
        }
        entity = create_vmexclude_model('12345user', 2)
        create_vmexclude(entity)
        self.task.info(None, 'mock', 'instance.create.end', create_msg, None)
        repo = repositories.get_vmexpire_repository()
        try:
            repo.get_by_instance(create_msg['nova_object.data']['uuid'])
        except Exception as e:
            logging.info('instance not found: ' + str(e))
        else:
            self.self.fail('domain is excluded, should not have been created')


def create_vmexpire_model(prefix=None):
    if not prefix:
        prefix = '12345'
    entity = models.VmExpire()
    entity.user_id = prefix + 'user'
    entity.project_id = prefix + 'project'
    entity.instance_id = prefix + 'instance'
    entity.instance_name = prefix
    entity.notified = False
    entity.notified_last = False
    entity.expire = time.mktime(datetime.datetime.now().timetuple())
    return entity


def create_vmexpire(entity):
    repo = repositories.get_vmexpire_repository()
    instance = repo.create_from(entity)
    repositories.commit()
    return instance


def create_vmexclude_model(exclude_id, exclude_type):
    entity = models.VmExclude()
    entity.exclude_id = exclude_id
    entity.exclude_type = exclude_type
    return entity


def create_vmexclude(entity):
    repo = repositories.get_vmexclude_repository()
    instance = repo.create_exclude(entity)
    repositories.commit()
    return instance



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

# from oslo_log import versionutils
import pecan
# import time
# import datetime

# from os_vm_expire import api
from os_vm_expire import api
from os_vm_expire.api import controllers
from os_vm_expire.common import hrefs
from os_vm_expire.common import utils
from os_vm_expire import i18n as u
# from os_vm_expire.model import models
from os_vm_expire.model import repositories as repo

from os_vm_expire.common import config


CONF = config.CONF

LOG = utils.getLogger(__name__)


def _vmexclude_not_found():
    """Throw exception indicating order not found."""
    pecan.abort(404, u._('Not Found. Sorry but your exclude is in '
                         'another castle.'))


class VmExcludeController(controllers.ACLMixin):

    """Handles Order retrieval and deletion requests."""

    def __init__(self, project_id):
        self.project_id = str(project_id)
        self.vmexclude_repo = repo.get_vmexclude_repository()

    @pecan.expose(generic=True)
    def index(self, **kwargs):
        pecan.abort(405)  # HTTP 405 Method Not Allowed as default

    @index.when(method='GET', template='json')
    @controllers.handle_exceptions(u._('VmExclude retrieval'))
    @controllers.enforce_rbac('vmexclude:get')
    def on_get(self, meta, instance_id=None):
        # if null get all else get exclude instance
        vm_repo = self.vmexclude_repo
        instances = []
        if instance_id is None:
            instances = vm_repo.get_type_entities()
        else:
            instance = vm_repo.get(entity_id=str(instance_id))
            # url = hrefs.convert_vmexpire_to_href(instance.id)
            repo.commit()
            return {
                'vmexclude': hrefs.convert_to_hrefs(instance.to_dict_fields())
                }

        total = len(instances)
        instances_resp = [
            hrefs.convert_to_hrefs(o.to_dict_fields())
            for o in instances
        ]
        instances_resp_overall = hrefs.add_nav_hrefs(
            'vmexcludes',
            0, total, total,
            {'vmexcludes': instances_resp}
            )
        instances_resp_overall = hrefs.add_self_href(self.project_id + '/vmexcludes/', instances_resp_overall)
        instances_resp_overall.update({'total': total})
        repo.commit()
        return instances_resp_overall

    @index.when(method='POST', template='json')
    @controllers.handle_exceptions(u._('VmExclude create'))
    @controllers.enforce_rbac('vmexclude:create')
    @controllers.enforce_content_types(['application/json'])
    def on_post(self, meta):
        instance = None
        data = api.load_body(pecan.request)
        entity_id = data.get('id')
        entity_type = data.get('type')
        if not entity_id or entity_type is None:
            pecan.response.status = 403
            return 'Empty id or type'
        if entity_type not in ['domain', 'project', 'user']:
            pecan.response.status = 403
            return 'Invalid type'
        else:
            entity_type = self.vmexclude_repo.get_exclude_type(entity_type)
        try:
            entity = self.vmexclude_repo.create_exclude_entity(entity_id, entity_type)
            instance = self.vmexclude_repo.create_exclude(entity)
        except Exception as e:
            pecan.response.status = 403
            return str(e)
        # url = hrefs.convert_vmexpire_to_href(instance.id)
        repo.commit()
        pecan.response.status = 202
        return {
            'vmexclude': hrefs.convert_to_hrefs(instance.to_dict_fields())
        }

    @index.when(method='DELETE', template='json')
    @controllers.handle_exceptions(u._('VmExclude deletion'))
    @controllers.enforce_rbac('vmexclude:delete')
    @controllers.enforce_content_types(['application/json'])
    def on_delete(self, meta, instance_id):
        instance = self.vmexclude_repo.get(entity_id=instance_id)
        if instance:
            self.vmexclude_repo.delete_entity_by_id(entity_id=instance.id)
            repo.commit()
        pecan.response.status = 204
        return



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import pecan
from six.moves.urllib import parse

from os_vm_expire.api import controllers
from os_vm_expire.api.controllers import vmexclude
from os_vm_expire.api.controllers import vmexpire
from os_vm_expire.common import utils
from os_vm_expire import i18n as u
from os_vm_expire import version

LOG = utils.getLogger(__name__)


MIME_TYPE_JSON = 'application/json'
MIME_TYPE_JSON_HOME = 'application/json-home'
MEDIA_TYPE_JSON = 'application/vnd.openstack.key-manager-%s+json'


def _version_not_found():
    """Throw exception indicating version not found."""
    pecan.abort(404, u._("The version you requested wasn't found"))


def _get_versioned_url(version):
    if version[-1] != '/':
        version += '/'
    # If host_href is not set in osvmexpire conf,
    # then derive it from request url
    host_part = utils.get_base_url_from_request()
    if host_part[-1] != '/':
        host_part += '/'
    return parse.urljoin(host_part, version)


class BaseVersionController(object):
    """Base class for the version-specific controllers"""

    @classmethod
    def get_version_info(cls, request):
        return {
            'id': cls.version_id,
            'status': 'stable',
            'updated': cls.last_updated,
            'links': [
                {
                    'rel': 'self',
                    'href': _get_versioned_url(cls.version_string),
                }, {
                    'rel': 'describedby',
                    'type': 'text/html',
                    'href': 'https://docs.openstack.org/'
                }
            ],
            'media-types': [
                {
                    'base': MIME_TYPE_JSON,
                    'type': MEDIA_TYPE_JSON % cls.version_string
                }
            ]
        }


class V1Controller(BaseVersionController):
    """Root controller for the v1 API"""

    version_string = 'v1'

    # NOTE(jaosorior): We might start using decimals in the future, meanwhile
    # this is the same as the version string.
    version_id = 'v1'

    last_updated = '2017-11-22T00:00:00Z'

    def __init__(self):
        LOG.debug('=== Creating V1Controller ===')

    @pecan.expose(generic=True)
    def index(self):
        pecan.abort(405)  # HTTP 405 Method Not Allowed as default

    @pecan.expose()
    def _lookup(self, project_id, *remainder):
        if not project_id:
            return self.on_get()
        if remainder and remainder[0] == 'vmexpires':
            return vmexpire.VmExpireController(project_id), remainder
        elif remainder and remainder[0] == 'vmexcludes':
            return vmexclude.VmExcludeController(project_id), remainder
        else:
            return self.on_get()

    @index.when(method='GET', template='json')
    @utils.allow_certain_content_types(MIME_TYPE_JSON, MIME_TYPE_JSON_HOME)
    @controllers.handle_exceptions(u._('Version retrieval'))
    def on_get(self):
        pecan.core.override_template('json')
        return {'version': self.get_version_info(pecan.request)}


AVAILABLE_VERSIONS = {
    V1Controller.version_string: V1Controller,
}

DEFAULT_VERSION = V1Controller.version_string


class VersionsController(object):

    def __init__(self):
        LOG.debug('=== Creating VersionsController ===')

    @pecan.expose(generic=True)
    def index(self, **kwargs):
        pecan.abort(405)  # HTTP 405 Method Not Allowed as default

    @index.when(method='GET', template='json')
    @utils.allow_certain_content_types(MIME_TYPE_JSON, MIME_TYPE_JSON_HOME)
    def on_get(self, **kwargs):
        """The list of versions is dependent on the context."""
        self._redirect_to_default_json_home_if_needed(pecan.request)

        if 'build' in kwargs:
            return {'build': version.__version__}

        versions_info = [version_class.get_version_info(pecan.request) for
                         version_class in AVAILABLE_VERSIONS.values()]

        version_output = {
            'versions': {
                'values': versions_info
            }
        }

        # Since we are returning all the versions available, the proper status
        # code is Multiple Choices (300)
        pecan.response.status = 300
        return version_output

    def _redirect_to_default_json_home_if_needed(self, request):
        if self._mime_best_match(request.accept) == MIME_TYPE_JSON_HOME:
            url = _get_versioned_url(DEFAULT_VERSION)
            LOG.debug("Redirecting Request to " + url)
            # NOTE(jaosorior): This issues an "external" redirect because of
            # two reasons:
            # * This module doesn't require authorization, and accessing
            #   specific version info needs that.
            # * The resource is a separate app_factory and won't be found
            #   internally
            pecan.redirect(url, request=request)

    def _mime_best_match(self, accept):
        if not accept:
            return MIME_TYPE_JSON

        SUPPORTED_TYPES = [MIME_TYPE_JSON, MIME_TYPE_JSON_HOME]
        return accept.best_match(SUPPORTED_TYPES)



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""create exclude table

Revision ID: 3cf9516e9a67
Revises: newton
Create Date: 2017-12-27 17:23:22.288802

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3cf9516e9a67'
down_revision = 'newton'


def upgrade():
    ctx = op.get_context()
    con = op.get_bind()
    table_exists = ctx.dialect.has_table(con.engine, 'vmexclude')
    if not table_exists:
        op.create_table(
            'vmexclude',
            sa.Column('id', sa.String(length=36), nullable=False),
            sa.Column('created_at', sa.DateTime(), nullable=False),
            sa.Column('updated_at', sa.DateTime(), nullable=False),
            sa.Column('deleted_at', sa.DateTime(), nullable=True),
            sa.Column('deleted', sa.Boolean(), nullable=False),
            sa.Column('exclude_id', sa.String(255), index=False, nullable=False),
            sa.Column('exclude_type', sa.Integer, index=False, nullable=False),
            sa.PrimaryKeyConstraint('id'),
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import mock
import time

# from os_vm_expire import context
from os_vm_expire.model import models
from os_vm_expire.model import repositories
from os_vm_expire.tests import utils


class WhenTestingVmExpiresResource(utils.OsVMExpireAPIBaseTestCase):

    def setUp(self):
        super(WhenTestingVmExpiresResource, self).setUp()
        self.max_vm_total_duration = repositories.CONF.max_vm_total_duration

    def tearDown(self):
        super(WhenTestingVmExpiresResource, self).tearDown()
        repo = repositories.get_vmexpire_repository()
        repo.delete_all_entities()
        repositories.commit()
        repositories.CONF.max_vm_total_duration = self.max_vm_total_duration

    def test_can_get_vmexpires(self):
        entity = create_vmexpire_model()
        create_vmexpire(entity)
        _get_resp = self.app.get('/' + entity.project_id + '/vmexpires/')
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexpires', _get_resp.json)
        self.assertEqual(len(_get_resp.json['vmexpires']), 1)
        self.assertEqual(
            _get_resp.json['vmexpires'][0]['instance_id'],
            entity.instance_id
            )

    def test_user_fails_all_tenants_vmexpires(self):
        entity_user = create_vmexpire_model()
        create_vmexpire(entity_user)
        entity = create_vmexpire_model(prefix='admin')
        create_vmexpire(entity)
        req = mock.MagicMock()
        req.get_param.return_value = None

        self.app.get('/' + entity.project_id + '/vmexpires/?all_tenants=1', status=403)

    def test_admin_can_get_own_vmexpires(self):
        entity_user = create_vmexpire_model()
        create_vmexpire(entity_user)
        entity = create_vmexpire_model(prefix='admin')
        create_vmexpire(entity)
        self.app.extra_environ = {
            'os_vm_expire.context': self._build_context(self.project_id, is_admin=True)
        }
        _get_resp = self.app.get('/' + entity.project_id + '/vmexpires/')
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexpires', _get_resp.json)
        self.assertEqual(len(_get_resp.json['vmexpires']), 1)
        self.assertEqual(
            _get_resp.json['vmexpires'][0]['instance_id'],
            entity.instance_id
            )

    def test_admin_can_get_all_vmexpires(self):
        entity_user = create_vmexpire_model()
        create_vmexpire(entity_user)
        entity = create_vmexpire_model(prefix='admin')
        create_vmexpire(entity)
        self.app.extra_environ = {
            'os_vm_expire.context': self._build_context(self.project_id, is_admin=True)
        }
        _get_resp = self.app.get('/' + entity.project_id + '/vmexpires/?all_tenants=1')
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexpires', _get_resp.json)
        self.assertEqual(len(_get_resp.json['vmexpires']), 2)

    def test_can_get_vmexpire(self):
        entity = create_vmexpire_model()
        instance = create_vmexpire(entity)
        _get_resp = self.app.get(
            '/' + entity.project_id + '/vmexpires/' + instance.id
            )
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexpire', _get_resp.json)
        self.assertEqual(
            _get_resp.json['vmexpire']['instance_id'],
            entity.instance_id)

    def test_can_extend_vmexpire(self):
        entity = create_vmexpire_model()
        instance = create_vmexpire(entity)
        _get_existing_resp = self.app.get(
            '/' + entity.project_id + '/vmexpires/' + instance.id
            )
        _get_resp = self.app.put(
            '/' + entity.project_id + '/vmexpires/' + instance.id,
            headers={'Content-Type': 'application/json'}
            )
        self.assertEqual(202, _get_resp.status_int)
        self.assertIn('vmexpire', _get_resp.json)
        self.assertEqual(
            _get_resp.json['vmexpire']['instance_id'],
            _get_existing_resp.json['vmexpire']['instance_id']
            )
        prev_expire = _get_existing_resp.json['vmexpire']['expire']
        new_expire = _get_resp.json['vmexpire']['expire']
        self.assertTrue(
            new_expire > prev_expire
            )

    def test_cannot_extend_max_reached(self):
        repositories.CONF.max_vm_total_duration = 3
        entity = create_vmexpire_model()
        instance = create_vmexpire(entity)
        self.app.get(
            '/' + entity.project_id + '/vmexpires/' + instance.id
            )
        self.app.put(
            '/' + entity.project_id + '/vmexpires/' + instance.id,
            headers={'Content-Type': 'application/json'},
            status=403
            )

    def test_extend_reset_notified_vmexpire(self):
        entity = create_vmexpire_model()
        instance = create_vmexpire(entity)
        instance.notified = True
        instance.save()
        repositories.commit()
        _get_existing_resp = self.app.get(
            '/' + entity.project_id + '/vmexpires/' + instance.id
            )
        _get_resp = self.app.put(
            '/' + entity.project_id + '/vmexpires/' + instance.id,
            headers={'Content-Type': 'application/json'}
            )
        self.assertEqual(202, _get_resp.status_int)
        self.assertFalse(_get_resp.json['vmexpire']['notified'])
        self.assertIn('vmexpire', _get_resp.json)
        self.assertEqual(
            _get_resp.json['vmexpire']['instance_id'],
            _get_existing_resp.json['vmexpire']['instance_id']
            )
        prev_expire = _get_existing_resp.json['vmexpire']['expire']
        new_expire = _get_resp.json['vmexpire']['expire']
        self.assertTrue(
            new_expire > prev_expire
            )

    def test_can_delete_vmexpire(self):
        entity = create_vmexpire_model()
        instance = create_vmexpire(entity)
        _get_resp = self.app.delete(
            '/' + entity.project_id + '/vmexpires/' + instance.id,
            headers={'Content-Type': 'application/json'}
            )
        self.assertEqual(204, _get_resp.status_int)
        _get_resp = self.app.get('/' + entity.project_id + '/vmexpires/')
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexpires', _get_resp.json)
        self.assertEqual(len(_get_resp.json['vmexpires']), 0)


def create_vmexpire_model(prefix=None):
    if not prefix:
        prefix = '12345'
    entity = models.VmExpire()
    entity.user_id = prefix + 'user'
    entity.project_id = prefix + 'project'
    entity.instance_id = prefix + 'instance'
    entity.instance_name = prefix
    entity.notified = False
    entity.notified_last = False
    entity.expire = time.mktime(datetime.datetime.now().timetuple())
    return entity


def create_vmexpire(entity):
    repo = repositories.get_vmexpire_repository()
    instance = repo.create_from(entity)
    repositories.commit()
    return instance
